# 2014. 12. 9 by Hans Roh hansroh@gmail.com

__version__ = "0.28.7.11"

version_info = tuple (map (lambda x: not x.isdigit () and x or int (x),  __version__.split (".")))
NAME = "Skitai/%s.%s" % version_info [:2]

import threading
import sys, os
import h2
from rs4 import deco, importer
from rs4.psutil import service
from rs4.attrdict import AttrDict
from aquests.protocols import dns
from aquests.dbapi import DB_PGSQL, DB_POSTGRESQL, DB_SQLITE3, DB_REDIS, DB_MONGODB
from aquests.protocols.smtp import composer
import tempfile
import getopt
from . import lifetime
from . import mounted

def test_client (*args, **kargs):
	from .testutil.launcher import Launcher
	return Launcher (*args, **kargs)

def websocket (varname, timeout = 60, onopen = None, onclose = None):
    from functools import wraps
    from skitai import was as the_was
    
    def decorator(f):
        @wraps(f)
        def wrapper (*args, **kwargs):            
            was = the_was._get ()            
            if not was.wshasevent ():
                return f (*args, **kwargs)
            if was.wsinit ():
                return was.wsconfig (1, timeout, [varname,])
            elif was.wsopened ():
            	return onopen and onopen () or ''
            elif was.wsclosed ():                    
                return onclose and onclose () or ''
        return wrapper
    return decorator

    	
HAS_ATILA = None

PROTO_HTTP = "http"
PROTO_HTTPS = "https"
PROTO_WS = "ws"
PROTO_WSS = "wss"
DJANGO = "django"

STA_REQFAIL = REQFAIL = -1
STA_UNSENT = UNSENT = 0
STA_TIMEOUT = TIMEOUT = 1
STA_NETERR = NETERR = 2
STA_NORMAL = NORMAL = 3

WEBSOCKET_SIMPLE = 1
WEBSOCKET_GROUPCHAT = 5

WS_SIMPLE = 1
WS_THREADSAFE = 6
WS_GROUPCHAT = 5
WS_NQ = 128

WS_EVT_INIT = "init"
WS_EVT_OPEN = "open"
WS_EVT_CLOSE = "close"
WS_EVT_NONE = None

WS_MSG_JSON = "json"
WS_MSG_XMLRPC = "xmlrpc"
WS_MSG_GRPC = "grpc"
WS_MSG_TEXT = "text"
WS_MSG_DEFAULT = "text"

WS_OPCODE_TEXT = 0x1
WS_OPCODE_BINARY = 0x2
WS_OPCODE_CLOSE = 0x8
WS_OPCODE_PING = 0x9
WS_OPCODE_PONG = 0xa

class _WASPool:
	def __init__ (self):
		self.__wasc = None
		self.__p = {}
		
	def __get_id (self):
		return id (threading.currentThread ())
	
	def __repr__ (self):
		return "<class skitai.WASPool at %x>" % id (self)
			
	def __getattr__ (self, attr):
		_was = self._get ()
		if not _was.in__dict__ ("app") and hasattr (_was, 'request'):
			# it will be called WSGI middlewares except Atila,
			# So request object not need
			del _was.request			
		return  getattr (_was, attr)
			
	def __setattr__ (self, attr, value):
		if attr.startswith ("_WASPool__"):
			self.__dict__[attr] = value
		else:	
			setattr (self.__wasc, attr, value)
			for _id in self.__p:
				setattr (self.__p [_id], attr, value)
	
	def __delattr__ (self, attr):
		delattr (self.__wasc, attr)
		for _id in self.__p:
			delattr (self.__p [_id], attr, value)
	
	def _start (self, wasc):
		self.__wasc = wasc
	
	def _del (self):
		_id = self.__get_id ()
		try:
			del self.__p [_id]
		except KeyError:
			pass

	def _get (self):
		_id = self.__get_id ()
		try:
			return self.__p [_id]
		except KeyError:
			_was = self.__wasc ()
			self.__p [_id] = _was
			return _was


was = _WASPool ()
def start_was (wasc):
	global was
	
	detect_atila ()
	was._start (wasc)

def detect_atila ():
	# for avoid recursive importing
	try:
		import atila
	except ImportError:
		pass
	else:
		global HAS_ATILA
		HAS_ATILA = atila.Atila
			

#------------------------------------------------
# Configure
#------------------------------------------------
dconf = {'mount': {"default": []}, 'clusters': {}, 'max_ages': {}, 'log_off': [], 'dns_protocol': 'tcp'}

def pref (preset = False):
	from .wsgi_apps import Config
	
	class Pref (AttrDict):
		def __init__ (self):
			super (Pref, self).__init__ (self)
			self.__dict__ ["mountables"] = [] 
			
		def mount (self, *args, **kargs):
			self.__dict__ ["mountables"].append ((args, kargs))
	
	d = Pref ()
	d.config = Config (preset)
	return d

PROCESS_NAME = None
def get_proc_title ():	
	global PROCESS_NAME
	
	if PROCESS_NAME is None:
		a, b = os.path.split (os.path.join (os.getcwd (), sys.argv [0]))
		script = b.split(".")[0]
				
		PROCESS_NAME =  "skitai/%s%s" % (
			os.path.basename (a),
			script != "app" and "-" + script or ''
		)
	return PROCESS_NAME	
	
def getswd ():
	return os.path.dirname (os.path.join (os.getcwd (), sys.argv [0]))

def is_devel ():
	return os.environ.get ('SKITAI_ENV') != "PRODUCTION"

def joinpath (*pathes):
	return os.path.normpath (os.path.join (getswd (), *pathes))
abspath = joinpath

Win32Service = None
def set_service (service_class):
	global Win32Service	
	Win32Service = service_class

def manual_gc (interval = 60.0):
	lifetime.manual_gc (interval)	
	
def set_worker_critical_point (cpu_percent = 90.0, continuous = 3, interval = 20):
	from .http_server import http_server
	from .https_server import https_server	
	
	http_server.critical_point_cpu_overload = https_server.critical_point_cpu_overload = cpu_percent
	http_server.critical_point_continuous = https_server.critical_point_continuous = continuous
	http_server.maintern_interval = https_server.maintern_interval = interval

def log_off (*path):		
	global dconf
	for each in path:
		dconf ['log_off'].append (each)	

def set_dns_protocol (protocol = 'tcp'):		
	global dconf
	dconf ['dns_protocol'] = protocol
	
def set_max_age (path, max_age):
	global dconf
	dconf ["max_ages"][path] = max_age

def set_max_rcache (objmax):
	global dconf
	dconf ["rcache_objmax"] = objmax

def set_keep_alive (timeout):	
	global dconf
	dconf ["keep_alive"] = timeout

def set_backend_keep_alive (timeout):	
	global dconf
	dconf ["backend_keep_alive"] = timeout

def set_proxy_keep_alive (channel = 60, tunnel = 600):	
	from .handlers import proxy

	proxy.PROXY_KEEP_ALIVE = channel
	proxy.PROXY_TUNNEL_KEEP_ALIVE = tunnel
		
def set_request_timeout (timeout):
	global dconf	
	dconf ["network_timeout"] = timeout
set_network_timeout = set_request_timeout

def set_was_class (was_class):
	global dconf	
	dconf ["wasc"] = was_class
	
def deflu (*key):
	if "models-keys" not in dconf:
		dconf ["models-keys"] = []
	if isinstance (key [0], (list, tuple)):
		key = list (key [0])
	dconf ["models-keys"].extend (key)
addlu = trackers = lukeys = deflu

def defgs (*key):
	if "states-keys" not in dconf:
		dconf ["states-keys"] = []
	if isinstance (key [0], (list, tuple)):
		key = list (key [0])
	dconf ["states-keys"].extend (key)
	
def maybe_django (wsgi_path, appname):
	if not isinstance (wsgi_path, str):
		return
	if appname != "application":
		return
	settings = os.path.join (os.path.dirname (wsgi_path), 'settings.py')
	if os.path.exists (settings):
		root = os.path.dirname (os.path.dirname (wsgi_path))
		sys.path.insert (0, root)		
		return root
	
def mount (point, target, appname = "app", pref = pref (True), host = "default", path = None):
	global dconf
	
	def init_app (modpath, pref):
		modinit = os.path.join (os.path.dirname (modpath), "__init__.py")
		if os.path.isfile (modinit):
			mod = importer.from_file ("temp", modinit)
			hasattr (mod, "bootstrap") and mod.bootstrap (pref)

	maybe_django (target, appname)
	if path:
		if isinstance (path, str):
			path = [path]
		path.reverse ()	
		for each in path:			
			sys.path.insert (0, abspath (each))
			
	if hasattr (target, "__file__"):
		target = (target, '__export__.py')
	
	if type (target) is tuple:
		module, appfile = target
		target = os.path.join (os.path.dirname (module.__file__), "export", "skitai", appfile)
			
	if type (target) is not str:
		# app instance, find app location
		target = os.path.normpath (os.path.join (os.getcwd (), sys.argv [0]))		
	else:
		if target [0] == "@":
			appname = None
		else:
			target = joinpath (target)
	
	if host not in dconf ['mount']:
		dconf ['mount'][host] = []

	if os.path.isdir (target) or not appname:
		dconf ['mount'][host].append ((point, target, None))
	else:		
		init_app (target, pref)
		dconf ['mount'][host].append ((point,  (target, appname), pref))
		
mount_django = mount

def enable_forward (forward_to = 443, port = 80, ip = ""):
	global dconf
	dconf ['fws_address'] = ip
	dconf ['fws_port'] = forward_to
	dconf ['fws_to'] = port

def enable_gateway (enable_auth = False, secure_key = None, realm = "Skitai API Gateway"):
	global dconf
	dconf ["enable_gw"] = True
	dconf ["gw_auth"] = enable_auth,
	dconf ["gw_realm"] = realm,
	dconf ["gw_secret_key"] = secure_key

def _get_django_settings (settings_path):
	import importlib
	import django
	
	ap = abspath (settings_path)
	django_main, settings_file = os.path.split (ap)
	django_root, django_main_dir = os.path.split (django_main)
	settings_mod = "{}.{}".format (django_main_dir, settings_file.split (".")[0])
	
	if not os.environ.get ("DJANGO_SETTINGS_MODULE"):		
		sys.path.insert (0, django_root)		
		os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings_mod)
		
	return importlib.import_module(settings_mod).DATABASES

def _alias_django (name, settings_path):
	dbsettings = _get_django_settings (settings_path)
	default = dbsettings ['default']
	if default ['ENGINE'].endswith ('sqlite3'):			
		return alias (name, DB_SQLITE3, default ['NAME'])
	
	if default ['ENGINE'].find ("postgresql") != -1:	
		if not default.get ("PORT"):
			default ["PORT"] = 5432
		if not default.get ("HOST"):
			default ["HOST"] = "127.0.0.1"
		if not default.get ("USER"):
			default ["USER"] = ""
		if not default.get ("PASSWORD"):
			default ["PASSWORD"] = ""
		return alias (name, DB_PGSQL, "%(HOST)s:%(PORT)s/%(NAME)s/%(USER)s/%(PASSWORD)s" % default)

def alias (name, ctype, members, role = "", source = "", ssl = False):
	from .rpc.cluster_manager import AccessPolicy
	global dconf
	
	if name [0] == "@":
		name = name [1:]
	if dconf ["clusters"].get (name):
		return name, dconf ["clusters"][name]
	
	if ctype == DJANGO:
		alias = _alias_django (name, members)
		if alias is None:
			raise SystemError ("Database engine is not compatible")
		return alias
	
	policy = AccessPolicy (role, source)
	args = (ctype, members, policy, ssl)
	dconf ["clusters"][name] = args
	return name, args
	
def enable_cachefs (memmax = 0, diskmax = 0, path = None):
	global dconf	
	dconf ["cachefs_memmax"] = memmax
	dconf ["cachefs_dir"] = path
	dconf ["cachefs_diskmax"] = diskmax	
				
def enable_proxy (unsecure_https = False):
	global dconf
	dconf ["proxy"] = True
	dconf ["proxy_unsecure_https"] = unsecure_https
	if os.name == "posix": 
		dconf ['dns_protocol'] = 'udp'

def enable_file_logging (path = None):
	global dconf
	dconf ['logpath'] = path
	
def enable_blacklist (path):
	global dconf
	dconf ["blacklist_dir"] = path

def enable_ssl (certfile, keyfile, passphrase):
	global dconf			
	dconf ["certfile"] = certfile
	dconf ["keyfile"] = keyfile
	dconf ["passphrase"] = passphrase

def get_varpath (name):
	name = name.split ("/", 1)[-1].replace (":", "-").replace (" ", "-")
	return os.name == "posix" and '/var/tmp/skitai/%s' % name or os.path.join (tempfile.gettempdir(), name)

def get_logpath (name):	
	name = name.split ("/", 1)[-1].replace (":", "-").replace (" ", "-")	
	return os.name == "posix" and '/var/log/skitai/%s' % name or os.path.join (tempfile.gettempdir(), name)

OPTLIST = None
SKITAI_LONGOPTS = ["skitai-profile", "skitai-memtrack"]

def argopt (sopt = "", lopt = []):		
	global OPTLIST
	
	if "d" in sopt:
		raise SystemError ("-d is used by skitai, please change")
	sopt += "d"
	lopt += SKITAI_LONGOPTS
	OPTLIST = (sopt, lopt)	
	argopt = getopt.getopt (sys.argv [1:], sopt, lopt)
	
	opts_ = []	
	for k, v in argopt [0]:
		if k == "-d":
			continue
		elif k.startswith ("---"):
			continue
		opts_.append ((k, v))
		
	aopt_ = [] 
	for arg in argopt [1]:
		if arg in ("start", "stop", "status", "restart"):		
			continue
		aopt_.append (arg)
	return opts_, aopt_	

def getsysopt (name, default):
	try: 
		return sys.argv [sys.argv.index ("---{}".format (name)) + 1]
	except ValueError:
		return default

def hassysopt (name):
	return "---{}".format (name) in sys.argv
		
def get_command ():
	global OPTLIST
	
	if OPTLIST is None:
		argopt_ = "-d" in sys.argv [1:] and [("-d", None)] or [], sys.argv [1:]
	else:			
		argopt_ = getopt.getopt (sys.argv [1:], *OPTLIST)
	
	cmd = None
	for cmd_ in ("start", "stop", "status", "restart"):
		if cmd_ in argopt_ [1]:
			cmd = cmd_
			break
	
	karg = {}
	for k, v in argopt_ [0]:
		if k == "-d": 
			cmd = "start"
			break		
	
	return cmd

def sched (interval, func): 
	lifetime.maintern.sched (interval, func)
					
def run (**conf):
	import os, sys, time	
	from . import Skitai
	from rs4.psutil import flock
	from rs4 import pathtool
	import getopt
			
	class SkitaiServer (Skitai.Loader):
		NAME = 'instance'
		
		def __init__ (self, conf):
			self.conf = conf			
			self.flock = None
			Skitai.Loader.__init__ (self, 'config', conf.get ('logpath'), conf.get ('varpath'), conf.get ("wasc"))
			
		def close (self):
			if self.wasc.httpserver.worker_ident == "master":
				pass
			Skitai.Loader.close (self)
			
		def config_logger (self, path):
			media = []
			if path is not None:
				media.append ("file")
			if self.conf.get ('verbose', "no") in ("yes", "1", 1):				
				media.append ("screen")
				self.conf ['verbose'] = "yes"
			if not media:
				media.append ("screen")
				self.conf ['verbose'] = "yes"
			Skitai.Loader.config_logger (self, path, media, self.conf ["log_off"])
		
		def master_jobs (self):
			self.wasc.logger ("server", "[info] engine tmp path: %s" % self.varpath)
			if self.logpath:
				self.wasc.logger ("server", "[info] engine log path: %s" % self.logpath)
			
			self.conf.get ("models-keys") and self.set_model_keys (self.conf ["models-keys"])
			self.conf.get ("states-keys") and self.set_state_keys (self.conf ["states-keys"])
						
		def maintern_shutdown_request (self, now):
			req = self.flock.lockread ("signal")
			if not req: return
			self.wasc.logger ("server", "[info] got signal - %s" % req)
			if req == "terminate":			
				lifetime.shutdown (0, 30.0)
			elif req == "restart":			
				lifetime.shutdown (3, 30.0)
			elif req == "kill":
				lifetime.shutdown (0, 1.0)
			elif req == "rotate":
				self.wasc.logger.rotate ()
			else:
				self.wasc.logger ("server", "[error] unknown signal - %s" % req)
			self.flock.unlock ("signal")
				
		def configure (self):
			conf = self.conf
			self.set_num_worker (conf.get ('workers', 1))
			if conf.get ("certfile"):
				self.config_certification (conf.get ("certfile"), conf.get ("keyfile"), conf.get ("passphrase"))
			
			self.config_dns (dconf ['dns_protocol'])
			if conf.get ("cachefs_diskmax", 0) and not conf.get ("cachefs_dir"):
				conf ["cachefs_dir"] = os.path.join (self.varpath, "cachefs")

			self.config_cachefs (
				conf.get ("cachefs_dir", None), 
				conf.get ("cachefs_memmax", 0),
				conf.get ("cachefs_diskmax", 0)				
			)
			self.config_rcache (conf.get ("rcache_objmax", 100))
			if conf.get ('fws_to'):
				self.config_forward_server (
					conf.get ('fws_address', '0.0.0.0'), conf.get ('fws_port', 80), conf.get ('fws_to', 443)
				)
			
			if "---port" in sys.argv:
				port = int (sys.argv [sys.argv.index ("---port") + 1])
			else:
				port = conf.get ('port', 5000)
								 	
			self.config_webserver (
				port, conf.get ('address', '0.0.0.0'),
				NAME, conf.get ("certfile") is not None,
				conf.get ('keep_alive', 2), 
				conf.get ('network_timeout', 10),
				thunks = [self.master_jobs]
			)
			
			if os.name == "posix" and self.wasc.httpserver.worker_ident == "master":
				# master does not serve
				return			
			
			self.config_threads (conf.get ('threads', 4))			
			self.config_backends (conf.get ('backend_keep_alive', 1200))
			for name, args in conf.get ("clusters", {}).items ():				
				ctype, members, policy, ssl = args
				self.add_cluster (ctype, name, members, ssl, policy)
			
			self.install_handler (
				conf.get ("mount"), 
				conf.get ("proxy", False),
				conf.get ("max_ages", {}),
				conf.get ("blacklist_dir"), # blacklist_dir
				conf.get ("proxy_unsecure_https", False), # disable unsecure https
				conf.get ("enable_gw", False), # API gateway
				conf.get ("gw_auth", False),
				conf.get ("gw_realm", "API Gateway"),
				conf.get ("gw_secret_key", None)
			)
			
			lifetime.init (logger = self.wasc.logger.get ("server"))
			lifetime.maintern.sched (3.0, dns.pool.maintern)
			if os.name == "nt":
				lifetime.maintern.sched (11.0, self.maintern_shutdown_request)								
				self.flock = flock.Lock (os.path.join (self.varpath, ".%s" % self.NAME))
	
	#----------------------------------------------------------------------------
	global dconf, PROCESS_NAME, Win32Service
	
	for k, v in dconf.items ():
		if k not in conf:
			conf [k] = v
	
	if "---production" in sys.argv:
		os.environ ["SKITAI_ENV"] = "PRODUCTION"
	
	if conf.get ("name"):
		PROCESS_NAME = 'skitai/{}'.format (conf ["name"])				
	if not conf.get ('mount'):
		raise systemError ('No mount point')
	conf ["varpath"] = get_varpath (get_proc_title ())
	pathtool.mkdir (conf ["varpath"])
	if "logpath" in conf and not conf ["logpath"]:
		conf ["logpath"] = get_logpath (get_proc_title ())
	
	cmd = get_command ()
	working_dir = getswd ()
	lockpath = conf ["varpath"]
	servicer = service.Service (get_proc_title(), working_dir, lockpath, Win32Service)

	if cmd and not servicer.execute (cmd):
		return
	if not cmd:
		if servicer.status (False):
			raise SystemError ("daemon is running")
		conf ['verbose'] = 'yes'
	elif cmd in ("start", "restart"):
		sys.stderr = open (os.path.join (conf.get ('varpath'), "stderr.engine"), "a")	
	
	server = SkitaiServer (conf)
	# timeout for fast keyboard interrupt on win32	
	try:
		try:
			server.run (conf.get ('verbose') and 1.0 or 30.0)
		except KeyboardInterrupt:
			pass	
	
	finally:	
		_exit_code = server.get_exit_code ()
		if _exit_code is not None: # master process				
			sys.exit (_exit_code)
		else: 
			# worker process				
			# for avoiding multiprocessing.manager process's join error
			os._exit (lifetime._exit_code)



import threading
from aquests.client import asynconnect
from aquests.client.socketpool import PROTO_CONCURRENT_STREAMS, select_channel
import time
import re
import copy
import random
from operator import itemgetter
import math
from urllib.parse import unquote

class AccessPolicy:    
    def __init__ (self, roles, ips):
        self.roles = self.to_list (roles)
        self.ips = []
        for ip_mask in self.to_list (ips):
            try: ip, mask = ip_mask.split ("/", 1)
            except ValueError: ip, mask = ip_mask, 32
            else: mask = int (mask)        
            pp = math.floor (mask / 8)
            mask_bits = mask % 8
            mask_bit = "1" * mask_bits
            mask_bit += "0" * (8 - mask_bits)
            self.ips.append ((ip.split (".")[:pp], mask, pp, int (mask_bit, 2)))
            
    def to_list (self, s):    
        return list (filter (None, map (lambda x: x.strip (), s.split (","))))
    
    def has_role (self, roles):
        if not self.roles:
            return roles and True or False
        for role in roles:
            if role in self.roles:
                return True
        return False
    
    def is_valid_request (self, request):
        if not self.ips:
            return True
        client_ip = request.channel.addr [0]
        for ip, mask, pp, mask_bit in self.ips:
            if mask == 32 and client_ip == ip:
                return True
            client_ip = client_ip.split (".")    
            if ip    != client_ip [:pp]:
                continue
            if mask_bit == 0:
                return True
            if int (client_ip [pp]) & mask_bit == mask_bit:
                return True
        return False        
            
            
class ClusterManager:
    object_timeout = 1200
    maintern_interval = 60
    # I cannot sure this is faster
    backend = True
    backend_keep_alive = 1200
    
    def __init__ (self, name, cluster, ssl = 0, access = None, logger = None):
        self.logger = logger
        self.lock = threading.RLock ()
        self._name = name
        self.access = access
        self.set_ssl (ssl)        
        self._proto = None
        self._havedeadnode = 0
        self._numget = 0
        self._nummget = 0
        self._clusterlist = []
        self._cluster = {}    
        self._last_maintern = time.time ()
        self._close_desires = []
        if cluster:
            self.create_pool (cluster)
    
    def __len__ (self):
        return len (self._cluster)
    
    def has_permission (self, request, roles):
        if self.access is None:
            return True
        return self.access.is_valid_request (request) and self.access.has_role (roles)
                
    def get_name (self):
        return self._name
            
    def is_ssl_cluster (self):
        return self._use_ssl
        
    def set_ssl (self, flag):
        self._use_ssl = flag
        if flag:
            self._conn_class = asynconnect.AsynSSLConnect    
        else:
            self._conn_class = asynconnect.AsynConnect
    
    def status (self):
        info = {}
        cluster = {}
        self.lock.acquire ()
        try:
            try:    
                for node in self._cluster:
                    _node = copy.copy (self._cluster [node])
                    _node ["numactives"] = len ([x for x in _node ["connection"] if x.isactive ()])
                    _node ["numconnected"] = len ([x for x in _node ["connection"] if x.isconnected ()])
                    
                    conns = []
                    for asyncon in _node ["connection"]:
                        conn = {
                                "connected": asyncon.isconnected (), 
                                "isactive": asyncon.isactive (), 
                                "request_count": asyncon.get_request_count (),
                                "event_time": time.asctime (time.localtime (asyncon.event_time)),
                                "zombie_timeout": asyncon.zombie_timeout,    
                                "keep_alive": asyncon.keep_alive,    
                            }
                        if hasattr (asyncon, "get_history"):
                            conn ["history"] = asyncon.get_history ()
                        conns.append (conn)
                                
                    _node ["connection"] = conns
                    cluster [str (node)] = _node
                info ["cluster"] = cluster
                
                actives = len ([x for x in self._close_desires if x.isactive ()])
                connecteds = len ([x for x in self._close_desires if x.isconnected ()])
                info ["close_pending"] = "%d (active: %d, connected: %d)" % (len (self._close_desires), actives, connecteds)
                info ["numget"] = self._numget
                info ["nummget"] = self._nummget
                info ["ssl"] = self.is_ssl_cluster ()                
                
            finally:
                self.lock.release ()
        except:
            self.logger.trace ()
        
        return info
        
    def cleanup (self):
        self.lock.acquire ()
        try:
            try:
                for node in list(self._cluster.keys ()):
                    self._cluster [node]["stoped"] = True
                    for asyncon in self._cluster [node]["connection"]:
                        asyncon.disconnect ()
            finally:    
                self.lock.release ()
        except:
            self.logger.trace ()    
    
    def get_nodes (self):
        nodes = []
        self.lock.acquire ()
        try:
            try:
                nodes = [k for k, v in list(self._cluster.items ()) if not v ["stoped"]]
            finally:    
                self.lock.release ()
        except:
            self.logger.trace ()                    
        return nodes
        
    def get_cluster (self):
        return self._cluster
    
    def parse_member (self, member):
        auth = None
        try:            
            userpass, netloc =  member.split ("@", 1)
        except ValueError:
            netloc =    member
        else:    
            try: 
                user, passwd = userpass.split (":", 1)
            except ValueError:
                user, passwd = userpass, ""
            auth  = (unquote (user), unquote (passwd))
        return auth, netloc
                    
    def create_asyncon (self, member):
        auth, netloc = self.parse_member (member)
        try: 
            host, port = netloc.split (":", 1)
            server = (host, int (port))
        except ValueError: 
            if not self._use_ssl:
                server    = (netloc, 80)            
            else:    
                server    = (netloc, 443)
        asyncon = self._conn_class (server, self.lock, self.logger)
        asyncon.set_auth (auth)
        self.backend and asyncon.set_backend (self.backend_keep_alive)
        return server, asyncon # nodeid, asyncon
        
    def add_node (self, member):
        try:                
            member, weight = member.split (" ", 1)
            weight = int (weight)
        except ValueError:
            weight = 1
        
        node, asyncon = self.create_asyncon (member)
        
        self.lock.acquire ()
        exists = node in self._cluster
        self.lock.release ()        
        if exists: 
            self._cluster [node]["weight"] = weight
            return
        
        _node = {"check": None, "connection": [], "weight": weight, "deadcount": 0, "stoped": False, "deadtotal": 0}
        _node ["connection"] = [asyncon]
                    
        self.lock.acquire ()        
        self._cluster [node] = _node
        self._clusterlist.append (node)
        self.lock.release ()
        
    def remove_node (self, member):
        host, port = member.split (":", 1)
        node = (host, int (port))
        self.lock.acquire ()
        try:
            try:    
                if node in self._cluster:
                    self._close_desires += self._cluster [node]["connection"]
                    del self._cluster [node]
                    del self._clusterlist [self._clusterlist.index (node)]                    
            finally:        
                self.lock.release ()
        except:
            self.logger.trace ()        
    
    def switch_node (self, member, stop = False):
        host, port = member.split (":", 1)
        node = (host, int (port))
        self.lock.acquire ()
        try:
            try:
                if node in self._cluster:
                    self._cluster [node]["stoped"] = stop
                    if stop is False:
                        self._cluster [node]["deadcount"] = 0
                        self._cluster [node]["check"] = None
                                        
            finally:        
                self.lock.release ()
        except:
            self.logger.trace ()
            
    def create_pool (self, cluster):
        for member in cluster:
            self.add_node (member)
            
    def report (self, asyncon, well_functioning):
        node = asyncon.address
        self.lock.acquire ()
        try:
            try:
                cluster = self._cluster
                if not well_functioning:
                    if cluster [node]["deadcount"] < 10:
                        recheck = 60
                    else:
                        recheck = 60 * 10
                    cluster [node]["check"] = time.time () + recheck
                    cluster [node]["deadcount"] += 1
                    cluster [node]["deadtotal"] += 1
                    
                else:
                    cluster [node]["deadcount"] = 0
                    cluster [node]["check"] = None
                    
            finally:    
                self.lock.release ()
                
        except:
            self.logger.trace ()        
    
    def maintern (self):
        try:
            # close unused sockets
            for _node in list(self._cluster.values ()):
                survived = []
                for asyncon in _node ["connection"]:
                    if not hasattr (asyncon, "maintern"):
                        continue
                                                                        
                    if asyncon.maintern (self.object_timeout):                    
                        asyncon.handler = None # break back ref.
                    else:                        
                        survived.append (asyncon)
                                                        
                if len (survived) == 0:
                    # at least 1 must be survived for duplicating
                    _node ["connection"] = _node ["connection"][:1]
                elif len (_node ["connection"]) != len (survived):
                    _node ["connection"] = survived
            
            # checking dead nodes
            if self._havedeadnode:
                for node in [k for k, v in list(self._cluster.items ()) if v ["check"] is not None and not v ["stoped"]]:
                    if time.time () > self._cluster [node]["check"]:
                        self._cluster [node]["check"] = None
            
            # closing removed mode's sockets                    
            if self._close_desires:
                cannot_closes = []
                for asyncon in self._close_desires:
                    if asyncon.isactive ():
                        cannot_closes.append (asyncon)
                    else:                        
                        asyncon.disconnect ()
                self._close_desires = cannot_closes
                
        except:
            self.logger.trace ()
        
    def get (self, specific = None, index = -1):
        asyncon = None
        try:
            with self.lock:
                self._numget += 1
                if time.time () - self._last_maintern > self.maintern_interval:
                    self._last_maintern = time.time ()
                    self.maintern ()                    
                    
                if specific:
                    nodes = [specific]
                
                elif index != -1:
                    try:
                        nodes = [self._clusterlist [index]]
                    except IndexError:
                        nodes = [self._clusterlist [-1]]
                        
                else:    
                    nodes = [k for k, v in list(self._cluster.items ()) if v ["check"] is None and not v ["stoped"]]
                    self._havedeadnode = len (self._cluster) - len (nodes)
                    if not nodes:
                        #assume all live...
                        nodes = [k for k, v in list(self._cluster.items ()) if not v ["stoped"]]
                                        
                cluster = []
                for node in nodes:
                    avails = [x for x in self._cluster [node]["connection"] if not x.isactive ()]
                    if self._proto and self._proto in PROTO_CONCURRENT_STREAMS:
                        # socket load-balancing
                        selected = select_channel (avails)                        
                        avail = selected and [selected] or []
                                                    
                    if not avails:
                        continue
                    
                    weight = self._cluster [node]["weight"]
                    actives = weight - len (avails)
                    
                    if actives == 0:
                        capability = 1.0
                    else:
                        capability = 1.0 - (actives / float (weight))
                    
                    avail = avails [0]
                    cluster.append ((avail, capability, weight))
                
                if cluster:
                    random.shuffle (cluster) # load balancing between same weighted members
                    cluster.sort (key = itemgetter(1, 2), reverse = True)
                    asyncon = cluster [0][0]
                    
                else:
                    t = [(len (self._cluster [node]["connection"]), node) for node in nodes]
                    t.sort ()
                    node = t [0][1]
                    asyncon = self._cluster [node]["connection"][0].duplicate ()
                    self._cluster [node]["connection"].append (asyncon)
                
                asyncon.set_active (True)
            
            if self._proto is None:
                self._proto = asyncon.get_proto ()
                
        except:
            self.logger.trace ()
                        
        return asyncon
    
    def sortfunc (self, a, b):
        r = b [1] - a [1]
        if r != 0: return r
        return b [2] - a [2]        
    
    def get_endpoints (self):
        endpoints = []
        scheme = self._use_ssl and "https" or "http"
        for node, config in self._cluster.items ():            
            port = node [1]
            ep = "{}://{}".format (scheme, node [0])            
            if not (scheme == "https" and port == 443 or scheme == "http" and port == 80):
                ep += ":{}".format (port)
            endpoints.append (ep)
        return endpoints   
            
            
if __name__ == "__main__":
    f = ClusterManager ('', [])
    print (f.parse_member ('asdada@www.myserver.com:5678/test'))
    


import sys
from ..utility import make_pushables
from ..exceptions import HTTPError
from ..rpc.cluster_dist_call import DEFAULT_TIMEOUT

class TaskBase:
    def __init__ (self, reqs, timeout = DEFAULT_TIMEOUT, cache = 0, cache_if = (200,)):
        assert isinstance (reqs, (list, tuple))        
        self.timeout = timeout        
        self.cache = cache
        self.cache_if = cache_if
        self.reqs = reqs        


class Tasks (TaskBase):
    def __init__ (self, reqs, timeout = DEFAULT_TIMEOUT, cache = 0, cache_if = (200,)):
        TaskBase.__init__ (self, reqs, timeout, cache, cache_if)
        self._results = []
        
    def __iter__ (self):
        return iter (self.results)
    
    def __getitem__ (self, sliced):
        return self.results [sliced]
                
    @property
    def results (self):       
        return self._results or self.dispatch ()
    
    def dispatch (self):
        self._results = [req.dispatch (self.timeout, self.cache, self.cache_if) for req in self.reqs]
        return self._results 
    
    def wait (self):
        self._results = [req.wait (self.timeout) for req in self.reqs]
        
        
class Futures (TaskBase):
    def __init__ (self, was, reqs, timeout = 10, cache = 0, cache_if = (200,)):
        TaskBase.__init__ (self, reqs, timeout, cache, cache_if)
        self._was = was
        self.args = {}
        self.fulfilled = None
        self.responded = 0        
        self.ress = [None] * len (self.reqs)
            
    def then (self, func, **kargs):
        self.args = kargs     
        self.fulfilled = func
        for reqid, req in enumerate (self.reqs):            
            req.set_callback (self._collect, reqid, self.timeout)
        return self
                 
    def _collect (self, res):
        self.responded += 1
        reqid = res.meta ["__reqid"]
        self.ress [reqid] = res
        self.cache and res.cache (self.cache, self.cache_if)
        if self.responded == len (self.reqs):
            if self.fulfilled:             
                self.respond ()
            else:
                self._was.response ("205 No Content", "")
                self._was.response.done ()
            
    def respond (self):
        response = self._was.response         
        try:            
            if self.args:
                content = self.fulfilled (self._was, self.ress, **self.args)
            else:
                content = self.fulfilled (self._was, self.ress)
            will_be_push = make_pushables (response, content)
            content = None
        except MemoryError:
            raise
        except HTTPError as e:
            response.start_response (e.status)
            content = response.build_error_template (e.explain, self._was)
        except:            
            self._was.traceback ()
            response.start_response ("502 Bad Gateway")
            content = response.build_error_template (self._was.app.debug and sys.exc_info () or None, self._was)            
       
        if content:
           will_be_push = make_pushables (response, content)
        
        if will_be_push is None:
            return
           
        for part in will_be_push:
            if len (will_be_push) == 1 and type (part) is bytes and len (response) == 0:
                response.update ("Content-Length", len (part))
            response.push (part)                
        response.done ()
    



import sqlite3
from psycopg2.pool import ThreadedConnectionPool
import redis
import pymongo
from skitai import DB_SQLITE3, DB_PGSQL, DB_REDIS, DB_MONGODB
import multiprocessing

CPU_COUNT = multiprocessing.cpu_count ()
PGPOOL= None

def make_endpoints (dbtype, from_list):     
    global PGPOOL, CPU_COUNT
       
    endpoints = []        
    for server, db, auth in from_list:
        user, password = "", ""
        if auth:
            if len (auth) == 2:
                user, password = auth
            else:
                user = auth [0]    
        if isinstance (server, str):
            try: 
                host, port = server.split (":", 1)
            except ValueError:
                host, port = server, None
        else:
             host, port = server
        
        kargs = {}
        if port: kargs ["port"] = port
        if password: kargs ["password"] = password
        
        if dbtype == DB_SQLITE3:
            conn = sqlite3.connect (host)
        elif dbtype == DB_PGSQL:
            if user: kargs ["user"] = user
            if PGPOOL is None:
                PGPOOL = ThreadedConnectionPool (0, CPU_COUNT * 3, host = host, database = db, **kargs)
            conn = PGPOOL.getconn ()
        elif dbtype == DB_REDIS:
            conn = redis.Redis (host = host, port = port, db = db)
        elif dbtype == DB_MONGODB:
            if user: kargs ["username"] = user
            conn = pymongo.MongoClient (host = host, **kargs)
        endpoints.append (conn)
    return endpoints

def restore (conns):
    for conn in conns:
        conn.close ()



import re, sys, os
try:
    from urllib.parse import unquote, unquote_plus
except ImportError:
    from urlparse import unquote
    from urllib import unquote_plus    
import json
from hashlib import md5
import random
from rs4 import producers
from collections import Iterator
from aquests.athreads import trigger

####################################################################################
# Utitlities
####################################################################################
REQUEST = re.compile ('([^ ]+) ([^ ]+)(( HTTP/([0-9.]+))$|$)')
CONNECTION = re.compile ('Connection: (.*)', re.IGNORECASE)

def crack_query (r):
    if type (r) is bytes:
        r = r.decode ("utf8")

    if not r: return {}
    if r[0]=='?': r=r[1:]    
    arg={}
    q = [x.split('=', 1) for x in r.split('&')]
    
    for each in q:
        k = unquote_plus (each[0])
        try: 
            t, k = k.split (":", 1)
        except ValueError:
            t = "str"
            
        if len (each) == 2:        
            v = unquote_plus(each[1])
            if t == "str":
                pass
            elif t == "int":
                v = int (v)
            elif t == "float":
                v = float (v)
            elif t == "list":
                v = v.split (",")
            elif t == "bit":
                v = v.lower () in ("1", "true", "yes")
            elif t == "json":
                v = json.loads (v)
                
        else:
            v = ""            
            if t == "str":
                pass
            elif t == "int":
                v = 0
            elif t == "float":
                v = 0.0
            elif t == "list":
                v = []
            elif t == "bit":
                v = False
            elif t == "json":
                v = {}
                
        if k in arg:
            if type (arg [k]) is not type ([]):
                arg[k] = [arg[k]]
            arg[k].append (v)
            
        else:
            arg[k] = v
            
    return arg
    
def crack_request (r):
    m = REQUEST.match (r)
    if m.end() == len(r):        
        uri=m.group(2)        
        if m.group(3):
            version = m.group(5)
        else:
            version = None    
        return m.group(1).lower(), uri, version
    else:
        return None, None, None

def join_headers (headers):
    r = []
    for i in range(len(headers)):
        if headers[i][0] in ' \t':    
            r[-1] = r[-1] + headers[i][1:]
        else:
            r.append (headers[i])
    return r

def get_header (head_reg, lines, group=1):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m.group (group)
    return ''

def get_header_match (head_reg, lines):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m
    return ''        

def get_extension (path):
    dirsep = path.rfind ('/')
    dotsep = path.rfind ('.')
    if dotsep > dirsep:
        return path[dotsep+1:]
    else:
        return ''

ALNUM = '0123456789abcdefghijklmnopqrstuvwxyz'
def md5uniqid (length = 13):    
    global ALNUM
    _id = ''
    for i in range (0, length):
        _id += random.choice(ALNUM)
    return md5 (_id.encode ("utf8")).hexdigest ()[length:]


def make_pushables (response, content):
    from .wastuff import futures
    from .wastuff.api import API
    
    if not response.is_responsable ():
        # already called response.done () or diconnected channel
        return
            
    if content is None: # Possibly no return mistake
        raise AssertionError ("Content or part should not be None")
    
    if not content: # explicit empty not content
        trigger.wakeup (lambda p=response: (p.done(),))
        return
            
    if type (content) not in (list, tuple):
        content = (content,) # make iterable
        
    if isinstance (content [0], futures.Futures):
        return
    
    if response ["content-type"] is None: 
        response ["Content-Type"] = "text/html"
                
    will_be_push = []
    if len (response) == 0:
        content_length = 0
    else:
        content_length = None
    
    for part in content:                
        # like Django response
        try: part = part.content
        except AttributeError: pass
            
        type_of_part = type (part)
        
        if isinstance (part, API):
            response.update ("Content-Type", part.get_content_type ())
            part = part.to_string ().encode ("utf8")
            will_be_push.append (part)
            
        elif type_of_part in (bytes, str):
            if len (part) == 0:
                continue
            if type_of_part is not bytes:
                try: 
                    part = part.encode ("utf8")
                except AttributeError:
                    raise AssertionError ("%s is not supportable content type" % str (type (part)).replace ("<", "&lt;").replace (">", "&gt;"))                                
                type_of_part = bytes
                
            if type_of_part is bytes:
                if content_length is not None:
                    content_length += len (part)
                will_be_push.append (part)
        
        else:    
            if hasattr (part, "read"):
                part = producers.closing_stream_producer (part)                
            elif type (part) is list:
                part = producers.list_producer (part)
            elif isinstance(part, Iterator): # flask etc.
                part = producers.iter_producer (part)
            
            if hasattr (part, "more"):
                # producer
                content_length = None
                # automatic close    when channel suddenly closed
                hasattr (part, "close") and response.die_with (part)
                # streaming obj
                hasattr (part, "ready") and response.set_streaming ()
                will_be_push.append (part)
            
            else:
                raise AssertionError ("Unsupoorted response object")
    
    if content_length is not None:
        response ["Content-Length"] = content_length
    
    return will_be_push    
    

def catch (format = 0, exc_info = None):
    if exc_info is None:
        exc_info = sys.exc_info()
    t, v, tb = exc_info
    tbinfo = []
    assert tb # Must have a traceback
    while tb:
        tbinfo.append((
            tb.tb_frame.f_code.co_filename,
            tb.tb_frame.f_code.co_name,
            str(tb.tb_lineno)
            ))
        tb = tb.tb_next

    del tb
    file, function, line = tbinfo [-1]
    
    # format 0 - text
    # format 1 - html
    # format 2 - list
    try:
        v = str (v)
    except:
        v = repr (v)
        
    if format == 1:
        if v.find ('\n') != -1:
            v = v.replace ("\r", "").replace ("\n", "<br>")
        buf = ["<h3>%s</h3><h4>%s</h4>" % (t.__name__.replace (">", "&gt;").replace ("<", "&lt;"), v)]
        buf.append ("<b>at %s at line %s, %s</b>" % (file, line, function == "?" and "__main__" or "function " + function))
        buf.append ("<ul type='square'>")
        buf += ["<li>%s <span class='f'>%s</span> <span class='n'><b>%s</b></font></li>" % x for x in tbinfo]
        buf.append ("</ul>")        
        return "\n".join (buf)

    buf = []
    buf.append ("%s %s" % (t, v))
    buf.append ("in file %s at line %s, %s" % (file, line, function == "?" and "__main__" or "function " + function))
    buf += ["%s %s %s" % x for x in tbinfo]
    if format == 0:
        return "\n".join (buf)
    return buf    
